#!/usr/bin/env python3

'Galactic Conquest for VisiData'

import sys
import math
import requests
import time
import getpass
import hashlib
import builtins

from visidata import *

option('refresh_rate_s', 2.0, 'time to sleep between refreshes')
option('disp_turn_done', '*', 'symbol to indicate player turn taken')

theme('disp_title', 'RCGameJam S1\'17', 'title on map sheet')
theme('color_dest_planet', 'underline', 'color of marked destination planet')
theme('color_empty_space', '20 blue', 'color of empty space')
theme('color_unowned_planet', 'white', 'color of unowned planet')
theme('disp_empty_space', '.', 'color of empty space')
theme('twinkle_rate', '200', 'neg twinkle')

command('G', 'vd.push(g_client.GameOptions)', 'push game options')
command('N', 'status(g_client.get("/regen_map")); g_client.Map.reload()', 'make New map')
command('Y', 'vd.push(g_client.Players)', 'push players sheet')
command('P', 'vd.push(g_client.Planets)', 'push planets sheet')
command('M', 'vd.push(g_client.Map)', 'push map sheet')
command('U', 'vd.push(g_client.QueuedDeployments)', 'push unsent routes sheet')
command('D', 'vd.push(g_client.HistoricalDeployments)', 'push historical deployments sheet')
command('E', 'vd.push(g_client.Events)', 'push events sheet')
command('R', 'vd.push(SheetList("scores", g_client.get("/scores").json()))', 'push scores sheet')
command('^S', 'g_client.submit_turn()', 'submit deployments and end turn')
command('Q', 'g_client.player_quit(); vd.sheets.clear()', 'quit the game (with confirm)')

options.disp_column_sep = ''
options.curses_timeout = '200'
#options.color_current_row = 'reverse white'

class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self


class WSIClient:
    def __init__(self, url):
        self.sessionid = None
        self.server_url = url

        self.Players = PlayersSheet()
        self.Planets = PlanetsSheet()
        self.QueuedDeployments = QueuedDeploymentsSheet()
        self.Events = EventsSheet()
        self.Map = MapSheet()
        self.HistoricalDeployments = HistoricalDeploymentsSheet()
        self.GameOptions = GameOptionsSheet()

        vd().rightStatus = self.rightStatus
        self.gamestate = {}
        self.refresh = False

    def submit_turn(self):
        if self.current_turn > 0:
            qd = self.QueuedDeployments
            for i, depl in enumerate(qd.rows):
                if depl.result:
                    continue
                try:
                    r = AttrDict(self.get('/deploy', **depl).json())
                    qd.rows[i] = r
                except Exception as e:
                    qd.rows[i].result = str(e)

            status(self.get('/end_turn').text)
        else:
            status(self.get("/ready").text)

    @property
    def current_turn(self):
        if not self.gamestate:
            self.refresh = True
            g_client.refresh_everything()

        return self.gamestate.get('current_turn')

    def rightStatus(self):
        max_turn = self.gamestate.get('num_turns')
        turn_num = self.current_turn

        rstatus = ''
        if turn_num:
            rstatus += 'Turn %s/%s:' % (turn_num, max_turn)

        for pl in self.Players.rows:
            name = pl.name
            if pl.ready:
                name += options.disp_turn_done
            if pl.name == self.username:
                fmtstr = ' [%s]'
            else:
                fmtstr = ' %s'
            rstatus += fmtstr % name

        return rstatus

    def login(self):  # before curses init
        self.username = builtins.input('player name: ')
        password = getpass.getpass('password: ')

        data = { 'username': self.username, 'password': hashlib.md5(password.encode('utf-8')).hexdigest() }

        r = requests.post(self.server_url + '/auth', data=data)

        if r.status_code == 200:
            self.sessionid = r.text
        else:
            error('[%s] %s' % (r.status_code, r.text))

        try:
            status(self.get('/join').text)
        except Exception as e:
            print(str(e))

    def get(self, path, **kwargs):
        if not self.sessionid:
            error('not logged in')

        kwargs['session'] = self.sessionid

        r = requests.get(self.server_url + path, params=kwargs)

        if r.status_code != 200:
            error('[%s] %s' % (r.status_code, r.text))

        return r

    def player_quit(self):
        yn = input('Are you sure you want to leave the game? (y/N) ')
        if yn[:1].upper() == 'Y':
            status(self.get('/player_quit').text)
        else:
            status('Whew! That was close.')

    def add_deployment(self, sources, dest, nships):
        ndeployments = 0
        for src in sources:
            r = self.get('/validate_deploy', launch_planet_name=src.name, dest_planet_name=dest.name, nships_requested=nships)
            if r.status_code != 200:
                status(r.text)
            else:
                d = AttrDict(r.json())
                d.result = None
                self.QueuedDeployments.rows.append(d)
                ndeployments += 1
                src.nships -= d.nships_requested  # keep track for the player
        status('Queued %s deployments' % ndeployments)

    @async
    def refresh_everything(self):
        while True:
            if self.refresh:
                self.Players.reload()  # to see who has taken their turn

                prev_turn = self.gamestate.get('current_turn', 0)
                self.gamestate = self.get('/gamestate').json()

                current_turn = self.gamestate.get('current_turn', 0)
                if current_turn != prev_turn:
                    self.Planets.reload()
                    self.Events.reload()
                    self.Map.reload()
                    self.HistoricalDeployments.reload()

                    if current_turn > self.gamestate.get('num_turns'):

                        scores = self.get("/scores").json()
                        status('Game over. %s is the winner!' % scores[0]['name'])
                        vd().push(SheetList("scores", scores))

                    elif current_turn > 0:
                        vd().push(self.Events)
                        status('Turn %s started' % current_turn)

            time.sleep(options.refresh_rate_s)


class PlayersSheet(Sheet):
    def __init__(self):
        super().__init__('players')

        self.columns = [
            ColumnAttr('name'),
            ColumnAttr('ready'),
        ]

        self.colorizers.append(lambda sheet,col,row,value: row and (row.color, 8))

    def reload(self):
        self.rows = [AttrDict(r) for r in g_client.get('/players').json()]
        self.rows.sort(key=lambda row: row.name)

    def get_player_color(self, playername):
        for plrow in self.rows:
            if plrow.name == playername:
                return plrow.color
        return options.color_unowned_planet

def ColumnItems(key_list, **kwargs):
    return [ColumnItem(x, x, **kwargs) for x in key_list]

def ColumnAttrs(key_list, **kwargs):
    return [ColumnAttr(x, **kwargs) for x in key_list]

def distance_turns(pl1, pl2):
    if pl1 and pl2:
        return max(int(math.sqrt((pl1.y-pl2.y)**2 + (pl1.x-pl2.x)**2)/2 + 0.5), 1)

class PlanetsSheet(Sheet):
    def __init__(self):
        super().__init__('planets')
        self.columns = [
            ColumnAttr('name'),
            ColumnAttr('prod', type=int),
            ColumnAttr('killpct', type=int, fmtstr='%s%%'),
            ColumnAttr('ownername'),
            ColumnAttr('nships', type=int),
            Column('turns_to_marked', type=int, getter=lambda row: distance_turns(row, self.marked_planet)),
            ColumnAttr('x', type=int, width=4),
            ColumnAttr('y', type=int, width=4),
        ]

        self.colorizers.append(lambda sheet,col,row,value: (g_client.Players.get_player_color(row.ownername), 5) if row else None)
        self.colorizers.append(lambda sheet,col,row,value: (options.color_dest_planet, 5) if row is sheet.marked_planet else None)
        self.marked_planet = None

        self.command(ENTER, 'vd.push(g_client.Map).cursorRowIndex = cursorRow.y; g_client.Map.cursorColIndex = cursorRow.x', 'go to this planet on the map')
        self.command('m', 'sheet.marked_planet = cursorRow', 'mark current planet as destination')
        self.command('f', 'g_client.add_deployment([cursorRow], marked_planet, int(input("# ships: ", value=cursorRow.nships)))', 'deploy N ships from current planet to marked planet')
        self.command('gf', 'g_client.add_deployment(selectedRows, marked_planet, int(input("# ships: ")))', 'deploy N ships from each selected planet to marked planet')

    def reload(self):
        # name, x, y, prod, killpct, owner.name, nships
        self.rows = [AttrDict(x) for x in g_client.get('/planets').json()]
        self.rows.sort(key=lambda row: row.name)


class QueuedDeploymentsSheet(Sheet):
    def __init__(self):
        super().__init__('queued_deployments')

        self.columns = [
            ColumnItem('src_turn', 'launch_turn', type=int),
            ColumnItem('dest_turn', 'dest_turn', type=int),
            ColumnItem('src', 'launch_planet_name'),
            ColumnItem('dest', 'dest_planet_name'),
            ColumnItem('nrequested', 'nships_requested', type=int),
            ColumnItem('ndeployed', 'nships_deployed', type=int),
            ColumnItem('result', 'result'),
        ]

        self.colorizers.append(self.colorIncomplete)

    @staticmethod
    def colorIncomplete(sheet, col, row, value):
        if row and col and col.name == 'ndeployed':
            if row.result and row.nships_deployed != row.nships_requested:
                return 'red bold', 5

    def reload(self):
        pass


class HistoricalDeploymentsSheet(Sheet):
    def __init__(self):
        super().__init__('deployments')
        self.columns = [
            ColumnItem('player', 'launch_player_name'),
            ColumnItem('src_turn', 'launch_turn', type=int),
            ColumnItem('dest_turn', 'dest_turn', type=int),
            ColumnItem('src', 'launch_planet_name'),
            ColumnItem('dest', 'dest_planet_name'),
            ColumnItem('nships', 'nships_deployed', type=int),
            ColumnItem('killpct', 'killpct', type=int, fmtstr='%s%%'),
        ]
        self.colorizers.append(self.colorDeployment)

    @staticmethod
    def colorDeployment(sheet, col, row, value):
        if row:
            return (g_client.Players.get_player_color(row['launch_player_name']), 9)

    def reload(self):
        self.rows = g_client.get('/deployments').json()


class EventsSheet(Sheet):
    def __init__(self):
        super().__init__('events')
        self.columns = [
            ColumnItem('turn', 'turn', type=int),
            ColumnItem('event', 'event', width=80),
        ]

    def reload(self):
        self.rows = g_client.get('/events').json()

        # find first event for the current turn and put the toprow there
        for i, (turn, _) in enumerate(self.rows):
            if turn == g_client.current_turn:
                self.topRowIndex = index
                self.cursorRowIndex = index
                break

class MapSheet(Sheet):
    def __init__(self):
        super().__init__('map')
        self.fieldToShow = [ 'name', 'prod', 'killpct', 'nships' ]
        self.command('m', 'g_client.Planets.marked_planet = cursorRow[cursorCol.x]', 'mark current planet as destination')
        self.command('f', 'g_client.add_deployment([cursorRow[cursorCol.x]], g_client.Planets.marked_planet, int(input("# ships: ", value=cursorRow[cursorCol.x].nships)))', 'deploy N ships from current planet to marked planet')
        self.command(' ', 'cycle_info()', 'cycle the information displayed')

        self.colorizers.append(self.colorOwnedPlanet)
        self.colorizers.append(self.colorRoguePlanet)
        self.colorizers.append(self.colorMarkedPlanet)
        self.colorizers.append(self.colorSpace)


    @property
    def title(self):
        return ''.join(c.name for c in columns)

    @title.setter
    def title(self, value):
        for i in range(self.map_w):
            self.columns[i].name = value[i:i+1]

    @staticmethod
    def colorMarkedPlanet(sheet,col,row,value):
        return (options.color_dest_planet, 5) if row and col and row[col.x] and row[col.x] is g_client.Planets.marked_planet else None

    @staticmethod
    def colorOwnedPlanet(sheet,col,row,value):
        return (g_client.Players.get_player_color(row[col.x].ownername), 9) if row and col and row[col.x] else None

    @staticmethod
    def colorRoguePlanet(sheet,col,row,value):
        return (g_client.Players.get_player_color(row[col.x].ownername), 9) if row and col and row[col.x] else None

    @staticmethod
    def colorSpace(sheet,col,row,value):
        if row and col and row[col.x] is None:
            if row is not sheet.cursorRow:
                r = options.color_empty_space
                if random.randrange(0, int(options.twinkle_rate)) == 0:
                    r = 'cyan bold'
                return (r, 2)

    def cycle_info(self):
        self.fieldToShow = self.fieldToShow[1:] + [self.fieldToShow[0]]
        status('showing "%s"' % self.fieldToShow[0])

    def reload(self):
        g_client.Planets.reload()

        # set columns on every reload, to use current width/height
        self.map_w = g_client.gamestate['map_width']
        self.map_h = g_client.gamestate['map_height']

        self.columns = []
        for x in range(self.map_w):
            c = Column(' ', width=3, getter=lambda row,x=x,self=self: getattr(row[x], self.fieldToShow[0]) or '?' if row[x] else options.disp_empty_space)
            c.x = x
            self.columns.append(c)
        self.title = options.disp_title

        # so the order can't be changed
        self.columns = tuple(self.columns)

        self.rows = []
        for y in range(self.map_h):
            current_row = []
            for x in range(self.map_w):
                current_row.append(None)
            self.rows.append(current_row)

        for planet in g_client.Planets.rows:
            self.rows[planet.y][planet.x] = planet

        self.rows = tuple(self.rows)

class GameOptionsSheet(Sheet):
    def __init__(self):
        super().__init__('game_options')
        self.columns = [
            ColumnItem('option', 0),
            ColumnItem('value', 1),
        ]
        self.command([ENTER, 'e'], 'status(g_client.get("/set_option", option=cursorRow[0], value=editCell(1))); reload()', 'edit game option')

    def reload(self):
        self.rows = list(g_client.get('/options').json().items())


if __name__ == '__main__':
    print(__doc__)

    server_addr = sys.argv[1] if len(sys.argv) > 1 else 'localhost'
    server_port = int(sys.argv[2] if len(sys.argv) > 2 else 80)

    g_client = WSIClient('http://%s:%s' % (server_addr, server_port))

    vd()._status = ["'N' to generate new 'M'ap; Ctrl-S when ready to start"]
    set_global('g_client', g_client)
    g_client.login()
    run([g_client.Players])




#!/usr/bin/env python3

from setuptools import setup

__version__ = '1.01'

setup(name='vdgalcon',
      version=__version__,
      install_requires='visidata requests'.split(),
      description='Galactic Conquest in curses, on VisiData',
      author='Saul Pwanson',
      author_email='vdgalcon@saul.pw',
      url='http://github.com/saulpw/vdgalcon',
      download_url='https://github.com/saulpw/vdgalcon/tarball/' + __version__,
      scripts=['vdgalcon.py', 'vdgalcon-server.py'],
      license='GPLv3',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Environment :: Console :: Curses',
          'Intended Audience :: Developers',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3',
          'Topic :: Office/Business :: Financial :: Spreadsheet',
      ],
      keywords=('console game textpunk visidata galactic conquest galcon')
      )



#!/usr/bin/env python3

import sys
import string
import math
import json
import http.server
import urllib.parse
import uuid
import random

player_colors = 'green yellow cyan magenta red blue'.split()
all_planet_names =  string.ascii_uppercase + '0123456789' + '☿♀♂♃♄⛢♅♆⚳⚴⚵' + '+&%$#@![](){}=<>'

rand = random.randrange

def error(s):
    raise HTTPException(300, s)

#### options management
class OptionsObject:
    """Get particular option value from `base_options`."""
    def __init__(self, d):
        self._opts = d
    def __getattr__(self, k):
        return self._opts[k]
    def __setitem__(self, k, v):
        if k not in self._opts:
            raise Exception('no such option "%s"' % k)
        self._opts[k] = v

class Game:
    def __init__(self):
        self.options_dict = {
            'num_planets': 57,
            'num_turns': 20,
            'map_width': 15,
            'map_height': 15,
            'toroidal_map': False,  # if the edges are connected
            'map_generator': 'rclogo_fixed',  # or 'random' or 'rclogo_var'
            'debug': False,
        }
        self.options = OptionsObject(self.options_dict)
        self.players = {}   # [playername] -> Player
        self.planets = {}   # [planetname] -> Planet
        self.deployments = []
        self.events = []
        self.current_turn = 0

        self.generate_planets()

    def generate_planets(self):
        f = globals()['generate_map_' + self.options.map_generator]

        self.planets = f(self.options.map_width,
                         self.options.map_height,
                         self.options.num_planets,
                         self.distance)

        # assign a starting planet for each player
        for i, player in enumerate(self.players.values()):
            starting_planet = self.planets[all_planet_names[i]]
            starting_planet.owner = player
            starting_planet.prod = 10
            starting_planet.nships = 10
            starting_planet.killpct = 40

        for planet in self.planets.values():
            if planet.name not in all_planet_names[:len(self.players)]:
                planet.prod = rand(6) + rand(6)
                planet.killpct = rand(11)+rand(11)+rand(11)+rand(11)+6
                planet.nships = max(planet.prod, 1)

    def notify(self, eventstr):
        self.events.append(Event(self.current_turn, eventstr))

    def start_game(self):
        self.next_turn()

    def end_turn(self):
        for d in self.deployments:
            if d.dest_turn != self.current_turn+1:
                continue

            if d.dest_planet.owner is d.launch_player:
                d.dest_planet.nships += d.nships_deployed
                self.notify('%s sent %d reinforcements to %s' % (d.launch_player, d.nships_deployed, d.dest_planet.name))
                if self.options.debug:
                    self.notify('   from %s' % d.launch_planet)
                continue

            # battle time

            attackers = d.nships_deployed
            defenders = d.dest_planet.nships
            attack_killpct = d.killpct
            defend_killpct = d.dest_planet.killpct

            battle_notice = '%s attacked %s (%s ships at %s%%) with %s ships (%s%%)' % (d.launch_player,
                d.dest_planet,
                defenders,
                defend_killpct,
                attackers,
                attack_killpct)

            round_results = []

            while attackers > 0 and defenders > 0:
                killed_by_attackers = sum(1 for i in range(attackers) if rand(100) <= attack_killpct)
                killed_by_defenders = sum(1 for i in range(defenders) if rand(100) <= defend_killpct)
                if self.options.debug:
                    round_results.append('%s/%s' % (
                        defenders-min(killed_by_attackers, defenders),
                        attackers-min(killed_by_defenders, attackers)))

                attackers -= killed_by_defenders
                defenders -= killed_by_attackers

            if round_results:
                battle_notice += '[%s]' % ','.join(round_results)

            if attackers <= defenders:
                defenders = max(defenders, 0)
                battle_notice += ' and was destroyed, leaving %s ships!' % defenders
                d.dest_planet.nships = defenders
            else:
                attackers = max(attackers, 0)
                battle_notice += ' and seized control with %s ships!' % attackers

                d.dest_planet.owner = d.launch_player
                d.dest_planet.nships = attackers

            self.notify(battle_notice)
            if self.options.debug:
                self.notify('   from %s' % d.launch_planet)

        for p in self.planets.values():
            if p.owner:
                # production
                p.nships += p.prod

                # random planetary events
                if rand(100) < 5:
                    if rand(40) > p.killpct:
                        p.killpct += rand(5)
                        self.notify('Planet %s instituted compulsory opera, killpct improved to %s%%' % (p.name, p.killpct))
                    elif rand(10) > p.prod:
                        p.prod += 1
                        self.notify('Planet %s improved its production to %s' % (p.name, p.prod))
                    elif rand(40) < p.killpct:
                        p.killpct -= rand(5)
                        self.notify('Planet %s legalized dancing, killpct dropped to %s%%' % (p.name, p.killpct))
                    elif rand(10) < p.prod:
                        p.prod -= 1
                        self.notify('Planet %s production decreased to %s%%' % (p.name, p.prod))
                    else:
                        self.GET_deploy(p.owner, p.name, random.choice([x.name for x in self.planets.values() if x.owner is p.owner]), None, p.nships)
                        self.notify('Planet %s revolted against the tyrannical rule of %s!' % (p.name, p.owner))
                        p.owner = None

        self.next_turn()

    def next_turn(self):
        self.current_turn += 1

        for p in self.players.values():
            p.turn_sent = False

        if self.current_turn > self.options.num_turns:
            scores = self.GET_scores(None)
            self.notify('Game over!')
            self.notify('%s is the winner!' % scores[0]['name'])
        else:
            self.notify('*** Turn %d started' % self.current_turn)

    def GET_scores(self, pl, **kwargs):
        player_scores = {}
        for plname in self.players.keys():
            player_scores[plname] = {
                'name': plname,
                'nplanets': 0,
                'nships': 0
            }

        for planet in self.planets.values():
            if planet.owner:
                player_scores[planet.owner.name]['nplanets'] += 1
                player_scores[planet.owner.name]['nships'] += planet.nships

        return sorted(player_scores.values(), key=lambda r: (r['nplanets'], r['nships']), reverse=True)

    def GET_player_quit(self, pl, **kwargs):
        if pl not in self.players.values():
            error('no such player in game')

        self.players.pop(pl.name)

        for planet in self.planets.values():
            if planet.owner is pl:
                planet.owner = None

        self.notify('Player %s has quit the game.' % pl.name)
        return 'Thanks for playing!'

    @property
    def started(self):
        return self.current_turn > 0

    def POST_options(self, pl, **kwargs):
        if self.started:
            return 'game already started'

        if pl not in self.players.values():
            return 'only players in the game can set the options'

        self.options.update(kwargs)

    def GET_options(self, pl, **kwargs):
        return self.options._opts

    def GET_set_option(self, pl, option='', value=''):
        if self.started:
            return 'options cannot be changed after game start'
        if pl.number is None:
            return 'only a participating player can change a game option'

        self.options[option] = type(getattr(self.options, option))(value)
        return 'options.%s is now %s' % (option, value)

    def GET_regen_map(self, pl, **kwargs):
        if self.started:
            error('game already started')

        self.generate_planets()

    def GET_gamestate(self, pl, **kwargs):
        return {
            'map_width': self.options.map_width,
            'map_height': self.options.map_height,
            'started': self.started,
            'current_turn': self.current_turn,
            'num_turns': self.options.num_turns,
        }

    # leaky
    def POST_auth(self, pl, **kwargs):
        return pl.sessionid

    def GET_join(self, pl, **kwargs):
        if self.started:
            raise HTTPException(402, 'Game already started')

        if pl.number is not None:
            return 'already player %s/%s' % (pl.number, len(self.players))

        pl.number = len(self.players)
        self.players[pl.name] = pl

        self.generate_planets()
        return 'joined game'

    def GET_ready(self, pl, **kwargs):
        if self.started:
            raise HTTPException(402, 'Game already started')

        if not pl:
            raise HTTPException(403, 'Unauthorized')

        pl.turn_sent = True
        if len(self.players) > 1 and all(pl.turn_sent for pl in self.players.values()):
            self.start_game()
            return 'game started'

        return 'player ready'

    def GET_players(self, pl, **kwargs):
        return [x.as_dict() for x in self.players.values()]

    def GET_planets(self, pl, **kwargs):
        return [x.as_dict() for x in self.planets.values()]

    def GET_deployments(self, pl, **kwargs):
        return [x.as_dict() for x in self.deployments if x.dest_turn <= self.current_turn or x.launch_player is pl]

    def GET_events(self, pl):
        return [x.as_dict() for x in self.events]

    def predeploy(self, launch_player, launch_planet_name=None, dest_planet_name=None, dest_turn=None, nships=0):
        if not self.started:
            error('game not yet started')

        launch_planet = self.planets.get(launch_planet_name) or error('no such planet %s' % launch_planet_name)
        if launch_player is not launch_planet.owner:
            error('player does not own planet')

        dest_planet = self.planets.get(dest_planet_name) or error('no such planet %s' % dest_planet_name)

        turns = int(self.distance(launch_planet, dest_planet)/2 + 0.5)
        turns = max(1, turns)

        if dest_turn is None:
            dest_turn = 0
        dest_turn = max(int(dest_turn), self.current_turn+turns)

        dobj = Deployment(launch_player, self.current_turn, launch_planet, dest_planet, dest_turn, int(nships), launch_planet.killpct)
        dobj.nships_deployed = min(int(nships), launch_planet.nships)
        return dobj

    def GET_deploy(self, launch_player, launch_planet_name=None, dest_planet_name=None, dest_turn=None, nships_requested=0, **kwargs):
        dobj = self.predeploy(launch_player, launch_planet_name, dest_planet_name, dest_turn, nships_requested)


        assert dobj.launch_planet.nships >= dobj.nships_deployed

        d = dobj.as_dict()

        if dobj.nships_deployed > 0:
            dobj.launch_planet.nships -= dobj.nships_deployed
            self.deployments.append(dobj)
            d['result'] = 'deployed'
        else:
            d['result'] = 'no ships'

        return d

    def GET_validate_deploy(self, launch_player, launch_planet_name=None, dest_planet_name=None, dest_turn=None, nships_requested=0, **kwargs):
        dobj = self.predeploy(launch_player, launch_planet_name, dest_planet_name, dest_turn, nships_requested)
        dobj.nships_deployed = 0  # for client to display correctly before actual deployment
        return dobj.as_dict()

    def GET_end_turn(self, pl):
        if not self.started:
            error('game not yet started')

        pl.turn_sent = True
        num_still_turning = len([p for p in self.players.values() if not p.turn_sent])
        if num_still_turning > 0:
            return 'still waiting on %d players' % num_still_turning
        else:
            return self.end_turn()

    def distance(self, here, dest):
        if self.options.toroidal_map:
            # toroidal distance : going off the edge comes back on the other side
            return math.sqrt( ((here.y-dest.y) % self.options.map_height)**2 + ((here.x-dest.x) % self.options.map_width)**2)
        else:
            return math.sqrt((here.y-dest.y)**2 + (here.x-dest.x)**2)


class Player:
    def __init__(self, name, md5_password, sessionid):
        self.number = None
        self.name = name
        self.md5_password = md5_password
        self.sessionid = sessionid
        self.turn_sent = False

    def as_dict(self):
        return dict(number=self.number,
                    name=self.name,
                    color=player_colors[self.number],
                    ready=self.turn_sent)

    def __str__(self):
        return self.name


class Planet:
    def __init__(self, name, x, y, prod=0, killpct=0, owner=None):
        self.name = name
        self.x = x
        self.y = y
        self.prod = prod
        self.killpct = killpct
        self.owner = owner
        self.nships = prod

    @property
    def xy(self):
        return (self.x, self.y)

    def __str__(self):
        return self.name

    def as_dict(self):
        r = {
                'name': self.name,
                'x': self.x,
                'y': self.y,
        }
        if self.owner:  # if owned by any player, all players can see it
            r.update(prod = self.prod,
                     killpct = self.killpct,
                     ownername = self.owner.name,
                     nships = self.nships)
        else:
            r.update(prod = None,
                     killpct = None,
                     ownername = None,
                     nships = None)

        return r

class Deployment:
    def __init__(self, launch_player, launch_turn, launch_planet, dest_planet, dest_turn, nships, killpct):
        self.launch_player = launch_player
        self.launch_turn = launch_turn
        self.launch_planet = launch_planet
        self.dest_planet = dest_planet
        self.dest_turn = dest_turn
        self.nships_requested = nships
        self.nships_deployed = 0
        self.killpct = killpct

    def as_dict(self):
        return {
            'launch_turn': self.launch_turn,
            'launch_player_name': self.launch_player.name,
            'launch_planet_name': self.launch_planet.name,
            'dest_planet_name': self.dest_planet.name,
            'dest_turn': int(self.dest_turn),
            'nships_requested': self.nships_requested,
            'nships_deployed': self.nships_deployed,
            'killpct': self.killpct
        }


class Event:
    def __init__(self, turn_num, eventstr):
        self.turn_num = turn_num
        self.eventstr = eventstr

    def as_dict(self):
        return {
            'turn': self.turn_num,
            'event': self.eventstr,
        }


### networking via simple HTTP

class HTTPException(Exception):
    def __init__(self, errcode, text):
        super().__init__(text)
        self.errcode = errcode


class WSIServer(http.server.HTTPServer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.game = Game()
        self.sessions = {}  # sessionid -> Player
        self.users = {}     # username -> Player


class WSIHandler(http.server.BaseHTTPRequestHandler):
    def generic_handler(self, reqtype, path, data):
        fields = urllib.parse.parse_qs(data)
        if 'params' in fields:
            fields = fields['params']

        fields = dict((k, v[0]) for k, v in fields.items())  # every param is a list; assume only one value for any param

        toplevel = path.split('/')[1]
        if toplevel:
            try:
                sessions = fields.pop('session', None)
                if sessions:
                    pl = self.server.sessions.get(sessions)
                else:
                    pl = None

                if not pl:
                    username = fields.get('username')
                    if username:
                        if username in self.server.users:
                            pl = self.server.users[username]
                            if fields['password'] != pl.md5_password:
                                raise HTTPException(403, 'Sorry, wrong password')
                        else:
                            sessionid = uuid.uuid1().hex
                            pl = Player(username, fields['password'], sessionid)
                            self.server.sessions[sessionid] = pl
                            self.server.users[username] = pl

                ret = getattr(self.server.game, '%s_%s' % (reqtype, toplevel))(pl, **fields)

                if isinstance(ret, list) or isinstance(ret, dict):
                    ret = json.dumps(ret)
                # else leave as string

                self.send_response(200)
                self.send_header('Content-type', 'text/json')
                self.end_headers()

                if ret:
                    self.wfile.write(ret.encode('utf-8'))
            except HTTPException as e:
                if self.server.game.options.debug:
                    import traceback
                    print(traceback.format_exc())
                self.send_response(e.errcode)
                self.send_header('Content-type', 'text/plain')
                self.end_headers()
                self.wfile.write(str(traceback.format_exc()).encode('utf-8')) #
            except Exception as e:
                import traceback
                print(traceback.format_exc())
                self.send_response(404)
                self.send_header('Content-type', 'text/plain')
                self.end_headers()
                self.wfile.write(str(traceback.format_exc()).encode('utf-8'))

    def do_GET(self):
        parsed_url = urllib.parse.urlparse(self.path)
        return self.generic_handler('GET', parsed_url.path, parsed_url.query)

    def do_POST(self):
        length = int(self.headers['content-length'])
        field_data = self.rfile.read(length).decode('utf-8')
        return self.generic_handler('POST', self.path, field_data)


def generate_map_random(width, height, num_planets, distancefunc):
    planets = {}

    for planet_name in all_planet_names[:num_planets]:
        x = rand(width)
        y = rand(height)
        while (x,y) in planet_coords:
            x = rand(width)
            y = rand(height)

        planets[planet_name] = Planet(planet_name, x, y)
        planet_coords.add((x,y))

    return planets


def generate_map_rclogo_fixed(width, height, num_planets, distancefunc):
    grid = '''
.X.X.X.G.X.X.X.
.X...........X.
....E.X.X......
.X...........X.
.....XX.XB.....
.X...........X.
...............
.C.X.X.H.X.X.D.
......X.X......
..X.X.X.X.X.X..
X.X.........X.X
....X.X.X.F....
X..A.X.X.X....X
...............
X.X.X.X.X.X.X.X'''

    starting_coords = {}
    rclogo_coords = []
    for y, line in enumerate(grid.strip().splitlines()):
        for x, ch in enumerate(line):
            if ch == 'X':
                rclogo_coords.append((x, y))
            elif ch != '.':
                starting_coords[ch] = (x, y)

    random.shuffle(rclogo_coords)

    planets = {}
    for planet_name in all_planet_names[:num_planets]:
        if planet_name in starting_coords:
            x, y = starting_coords[planet_name]
        else:
            x, y = rclogo_coords.pop()

        planets[planet_name] = Planet(planet_name, x, y)

    return planets


def generate_map_rclogo_var(width, height, num_planets, distancefunc):
    planet_names = all_planet_names[:num_planets]

    def allowed_coord_set(width, height):
        def rectangle(leftupper, rightlower): # inclusive
            return set([(i,j) for i in range(leftupper[0], rightlower[0]+1) for j in range(leftupper[1], rightlower[1]+1)])

        allowed = set()

        allowed = allowed | rectangle((2,2)                  ,(width-3,2))
        allowed = allowed | rectangle((2,int(2/3 * height)-2)  ,(width-3,int(2/3 * height)-2))
        allowed = allowed | rectangle((int(6/15 * width),int(2/3 * height)-1),(int(6/15 * width),int(2/3 * height))) # left stand
        allowed = allowed | rectangle((int(7/15 * width),int(2/3 * height)+1), (int(7/15*width), int(2/3 * height)+2)) # right stand
        allowed = allowed | rectangle((int(8/15 * width),int(2/3 * height)-1),(int(8/15 * width),int(2/3 * height))) # left stand
        allowed = allowed | rectangle((int(9/15 * width),int(2/3 * height)+1), (int(9/15*width), int(2/3 * height)+2)) # right stand
        allowed = allowed | rectangle( (int(width*2/10),height-3)        ,(int(3/10*width),height-3)) # left keyboard
        allowed = allowed | rectangle( (int(width*4/10),height-2)        ,(int(5/10*width),height-2)) # left keyboard
        allowed = allowed | rectangle( (int(width*6/10),height-3)        ,(int(7/10*width),height-3)) # left keyboard
        allowed = allowed | rectangle( (int(width*7/10),height-2)        ,(int(8/10*width),height-2)) # right keyboard

        allowed = allowed | rectangle( (2,2)                  ,(2,int(2/3 * height)-2) )  # left edge
        allowed = allowed | rectangle((width-3,2)             ,(width-3,int(2/3 * height)-2))  # right edge
        assert height not in [x[1] for x in allowed], "allowed set is too tall"
        assert width not in [x[0] for x in allowed], "allowed set is too wide"

        return allowed

    def planet_away_from_planets(width, height, existingplanets):
        def min_distance(oneplanet, planets):
            return min(map(lambda x, oneplanet=oneplanet, distancefunc=distancefunc: distancefunc(oneplanet,x), planets))

        def index_best(potentialplanets, ownedplanets):
            distances = map(lambda x, ownedplanets=ownedplanets: min_distance(x ,ownedplanets), potentialplanets)
            index_of_best_planet = max([(d,i) for i,d in enumerate(distances)])[1]
            return index_of_best_planet

        ownedplanets = [p for p in existingplanets.values() if p.owner is not None]
        potentialplanets = random.sample(list(existingplanets.values()),k=5)
        if len(ownedplanets) > 0 :
            idx =  index_best(ownedplanets, potentialplanets)
            return potentialplanets[idx]
        else:
            return potentialplanets[0]

    # body
#    if planets is not None:
#        owners = [p.owner for p in planets.values() if p.owner is not None]
#    else:
#        owners = []

#    owned_planet_coords = set([p.xy for p in planets.values() if p.owner is not None])
    newcoord_list = random.sample(allowed_coord_set(width, height), k=len(planet_names))

    planets = {}
    for i, planet_name in enumerate(planet_names[:num_planets]):
        planets[planet_name] = Planet(planet_name, newcoord_list[i][0], newcoord_list[i][1])

#    for i, (name, pl) in enumerate(self.players.items()):
#        if pl not in owners or planets is None  :
#            planet_name = planet_names[i]
#            (xx,yy) = planet_away_from_planets(width, height, self.planets).xy
#            self.planets[planet_name] = Planet(planet_name, xx, yy, 10, 40, pl)

    return planets


def main():
    import argparse
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('-p', dest='port', help='port to run on', default=8080)
    args = parser.parse_args()

    port = int(args.port)

    server = WSIServer(('', port), WSIHandler)

    import requests
    public_ip = requests.get('http://ip.42.pl/raw').text

    print('http://%s:%s' % (public_ip, port))
    server.serve_forever()


if __name__ == '__main__':
    main()

